from iso_4217 import Currency, __version__
from iso_4217.lists import ApproxDate, ApproxTimeSpan, Historic


def test_version():
    assert __version__ == "0.7.250512"


def test_currency_count():
    assert len(Currency) == 307


def test_active_currencies_count():
    assert sum(bool(c.entities) for c in Currency) == 179


def test_uah():
    assert Currency.UAH.name == "UAH"
    assert Currency.UAH.value == "Hryvnia"
    assert Currency.UAH.number == 980
    assert Currency.UAH.entities == frozenset({"UKRAINE"})
    assert Currency.UAH.withdrew_entities == ()
    assert Currency.UAH.is_fund is False
    assert Currency.UAH.subunit_exp == 2


def test_eur():
    assert Currency.EUR.name == "EUR"
    assert Currency.EUR.value == "Euro"
    assert Currency.EUR.number == 978
    assert len(Currency.EUR.entities) == 36
    assert Currency.EUR.withdrew_entities == (
        ("SERBIA AND MONTENEGRO", "Euro", ApproxTimeSpan(ApproxDate(2006, 10))),
    )
    assert Currency.EUR.is_fund is False
    assert Currency.EUR.subunit_exp == 2


def test_ron():
    assert Currency.RON.name == "RON"
    assert Currency.RON.value == "Romanian Leu"
    assert Currency.RON.number == 946
    assert Currency.RON.entities == frozenset({"ROMANIA"})
    assert Currency.RON.withdrew_entities == (
        ("ROMANIA", "New Romanian Leu", ApproxTimeSpan(ApproxDate(2015, 6))),
    )
    assert Currency.RON.is_fund is False
    assert Currency.RON.subunit_exp == 2


def test_zwg():
    assert Currency.ZWG.name == "ZWG"
    assert Currency.ZWG.value == "Zimbabwe Gold"
    assert Currency.ZWG.number == 924
    assert Currency.ZWG.entities == frozenset({"ZIMBABWE"})


def test_bgk_bgj():
    assert Currency.BGK.value == "Lev A/62 (1989)"
    assert Currency.BGJ.value == "Lev A/52 (1989)"


def test_missing_codes():
    assert Currency.XFO.value == "Gold-Franc (2006)"
    assert Currency.XRE.value == "RINET Funds Code (1999)"
    assert Currency.XFU.value == "UIC-Franc (2013)"


def test_ves_ved():
    assert Currency.VES.number == 928
    assert Currency.VES.value == "Bolívar Soberano (VES)"
    assert Currency.VED.number == 926
    assert Currency.VED.value == "Bolívar Soberano (VED)"


def test_zwr():
    assert Currency.ZWR.number == 935
    assert Currency.ZWR.entities == frozenset()
    assert Currency.ZWR.withdrew_entities == (
        Historic(
            "ZIMBABWE",
            "Zimbabwe Dollar",
            ApproxTimeSpan(end=ApproxDate(year=2009, month=6)),
        ),
    )



import enum
from typing import Dict, FrozenSet, Optional, Tuple

from .lists import Historic, load
from .pint import currency_unit

__published_date__, _TABLE = load()


def _generate_enum(locals: Dict) -> None:
    from collections import Counter

    # VES and VED currencies got the same name, should generate distinct values for them
    reused_names = Counter(c.name for c in _TABLE.values() if c.entities)
    reused_names = {name for name, count in reused_names.most_common() if count > 1}
    for code, currency in _TABLE.items():
        if currency.entities:
            if currency.name in reused_names:
                locals[code] = "{} ({})".format(currency.name, code)
            else:
                locals[code] = currency.name
        else:
            last_year = currency.withdrew_entities[-1].time.end.year
            locals[code] = "{} ({})".format(currency.name, last_year)


class Currency(enum.Enum):
    """
    Represents distinct currency found in "Current currency & funds code"
    and "Codes for historic denominations of currencies & funds" lists.

    *Member name* is a 3-letter uppercase currency code
    *Member value* is currency name. Historical currencies have withdrawl year added
                   to the name to differentiate currencies that were withdrawn multiple
                   times.

    >>> Currency['EUR'].name
    'EUR'
    >>> Currency['EUR'].number
    978
    >>> Currency['XFU'].value
    'UIC-Franc (2013)'
    >>> Currency.UYU
    <Currency.UYU: 'Peso Uruguayo'>
    >>> Currency.UYN
    <Currency.UYN: 'Old Uruguay Peso (1989)'>
    """

    @property
    def number(self) -> Optional[int]:
        """
        Three-digit currency code number.
        Many historical currencies share same number.

        >>> Currency.USD.number
        840
        >>> Currency.UYN.number
        858
        """
        return _TABLE[self.name].number

    @property
    def entities(self) -> FrozenSet[str]:
        """
        Set of countries or other entities using the currency.
        Empty for historical currencies.
        :return: frozen set of entity names.

        >>> Currency.RUB.entities
        frozenset({'RUSSIAN FEDERATION (THE)'})
        >>> Currency.ECV.entities
        frozenset()
        """
        return _TABLE[self.name].entities

    @property
    def withdrew_entities(self) -> Tuple[Historic, ...]:
        """
        Countries or other entities that do not use the currency anymore.
        :return: List of tuples of entities and dates when withdrawl happened.

        >>> Currency.UAK.withdrew_entities
        (Historic(entity='UKRAINE', name='Karbovanet', time=...),)
        """
        return _TABLE[self.name].withdrew_entities

    @property
    def is_fund(self) -> bool:
        """
        Currencies marked as funds.

        >>> Currency.USN.is_fund
        True
        """
        return _TABLE[self.name].is_fund

    @property
    def subunit_exp(self) -> Optional[int]:
        """
        Power of ten to produce amount of subunits in single currency unit
        :return: Number of digits after decimal point

        >>> Currency.BHD.subunit_exp
        3
        >>> Currency.USD.subunit_exp
        2
        >>> Currency.JPY.subunit_exp
        0
        """
        return _TABLE[self.name].subunit_exp

    @property
    def unit(self):
        """
        If the pint package is installed and currency units are defined by calling
        define_currency_units on any registry, this will return corresponding pint unit.

        >>> Currency.EUR.unit
        <Unit('EUR')>
        """
        return currency_unit(self.name)

    @property
    def subunit(self):
        """
        If the pint package is installed and currency units are defined by calling
        define_currency_units on any registry, this will return corresponding
        dummy pint subunit. These have non-standard currency codes, suffixed with 's'.

        >>> Currency.EUR.subunit
        <Unit('EURs')>
        """
        return currency_unit(self.name + "s")

    def __str__(self):
        return self.name

    _generate_enum(locals())



from .currency import Currency, __published_date__
from .pint import define_currency_units

__version_prefix__ = "0.7"
__version__ = "{}.{:%y%m%d}".format(__version_prefix__, __published_date__)
__all__ = ("Currency", "define_currency_units")



from decimal import Decimal

import pint
import pytest

from iso_4217 import define_currency_units


@pytest.fixture(scope="module")
def reg():
    reg = pint.UnitRegistry(non_int_type=Decimal)
    return define_currency_units(reg)


def test_hryvnias(reg):
    assert reg("5 Hryvnias") == 5 * reg.UAH


def test_us_dollars(reg):
    assert reg("5 US_Dollar") == 5 * reg.USD


def test_us_dollar_cents(reg):
    assert 5 * reg.USDs == Decimal("0.05") * reg.USD


def test_cant_convert_dollars_euro(reg):
    with pytest.raises(pint.DimensionalityError):
        (reg.USD * 5).to("EUR")


def test_convert_with_context_dollars_euro(reg):
    c = pint.Context()
    c.add_transformation(
        "[currency_USD]", "[currency_EUR]", lambda r, u: 1.2 * r.EUR / r.USD * u
    )
    dollars = 5 * reg.USD
    euros = 6 * reg.EUR

    assert dollars.to("EUR", c) == euros



from urllib.request import urlopen

import pytest

from iso_4217.lists import _load_xml_resource


@pytest.mark.parametrize("list_xml", ("list-one.xml", "list-three.xml"))
def test_data_tables_are_up_to_date(list_xml):
    LISTS_URL = (
        "https://www.six-group.com/dam/download/financial-information/"
        "data-center/iso-currrency/lists/"
    )
    assert _load_xml_resource(list_xml) == urlopen(LISTS_URL + list_xml).read()



from . import currency

try:
    from pint import UndefinedUnitError, get_application_registry
except ImportError:

    def get_application_registry():
        raise RuntimeError("Units support requires pint package installed.")


def define_currency_units(registry):
    """Define currency units in a given pint UnitRegistry"""
    for c in currency.Currency:
        if c.subunit_exp is None:
            registry.define("{0} = [currency_{0}]".format(c.name))
        else:
            alias1 = c.value.replace(" ", "_")
            alias2 = alias1.lower()
            registry.define(
                "{0} = [currency_{0}] = _ = {1} = {2}".format(c.name, alias1, alias2)
            )
            registry.define("{0}s = {0} / {1}".format(c.name, 10 ** c.subunit_exp))
    return registry


def currency_unit(name: str):
    registry = get_application_registry()
    try:
        return registry.Unit(name)
    except UndefinedUnitError:
        return define_currency_units(registry).Unit(name)



import locale
from datetime import datetime
from itertools import chain, groupby
from typing import (
    Callable,
    Dict,
    FrozenSet,
    Iterable,
    List,
    NamedTuple,
    Optional,
    Tuple,
)
from xml.etree import ElementTree


class ApproxDate(NamedTuple):
    """Parsed approximate currency withdrawal date"""

    year: int
    month: Optional[int] = None

    def __str__(self):
        year, month = self.year, self.month
        return "{}-{:02}".format(year, month) if month else str(year)

    @classmethod
    def from_str(cls, text: str):
        return cls(*(int(p) for p in text.split("-")))


class ApproxTimeSpan(NamedTuple):
    """Parsed approximate currency withdrawal time span"""

    end: ApproxDate
    begin: Optional[ApproxDate] = None

    def __str__(self):
        return "{} to {}".format(self.begin, self.end) if self.begin else str(self.end)

    @classmethod
    def from_str(cls, text: str):
        return cls(*(ApproxDate.from_str(date) for date in text.split(" to ")))


class Historic(NamedTuple):
    """
    Contains information about currency withdrawal event.
    """

    entity: str
    name: str
    time: ApproxTimeSpan


class CurrencyInfo(NamedTuple):
    """
    Contains all the information about currency joined from both lists
    """

    code: str
    name: str
    number: Optional[int]
    subunit_exp: Optional[int]
    is_fund: bool
    entities: FrozenSet[str]
    withdrew_entities: Tuple[Historic, ...]


def load() -> Tuple[datetime, dict]:
    current = locale.getlocale(locale.LC_TIME)
    locale.setlocale(locale.LC_TIME, "C")
    date1, active = _load_list("list-one.xml", "CcyTbl/CcyNtry", _currency_data)
    date2, historic = _load_list(
        "list-three.xml", "HstrcCcyTbl/HstrcCcyNtry", _historic_data
    )
    locale.setlocale(locale.LC_TIME, current)
    both = sorted(chain(active, historic), key=lambda c: c["code"])
    currencies = (
        _group_entities(list(g)) for _, g in groupby(both, lambda c: c["code"])
    )
    table = {c.code: c for c in currencies}
    return max(date1, date2), table


def _load_list(
    filename: str, path: str, convert: Callable[[ElementTree.Element], Optional[Dict]]
) -> Tuple[datetime, Iterable[Dict]]:
    tree = ElementTree.fromstring(_load_xml_resource(filename))
    date = datetime.strptime(tree.attrib["Pblshd"], "%Y-%m-%d")
    return date, filter(None, (convert(node) for node in tree.findall(path)))


def _load_xml_resource(filename: str) -> bytes:
    try:
        from importlib.resources import files

        return files(__package__).joinpath("data/" + filename).read_bytes()
    # Old python versions might miss files function or resources modules
    except (AttributeError, ImportError, TypeError):
        from pkg_resources import resource_string

        return resource_string(__name__, "data/" + filename)


def _currency_data(node: ElementTree.Element) -> Optional[Dict]:
    name = node.find("CcyNm")
    if name.text != "No universal currency":
        subunit_exp = node.find("CcyMnrUnts").text
        return dict(
            code=node.find("Ccy").text,
            number=int(node.find("CcyNbr").text),
            name=name.text.rstrip(),
            subunit_exp=int(subunit_exp) if subunit_exp != "N.A." else None,
            is_fund="IsFund" in name.attrib,
            entity=node.find("CtryNm").text.strip(),
        )


def _historic_data(node: ElementTree.Element) -> Dict:
    name = node.find("CcyNm")
    number = node.find("CcyNbr")
    return dict(
        code=node.find("Ccy").text,
        number=int(number.text) if number is not None else None,
        name=name.text.rstrip(),
        subunit_exp=None,
        is_fund="IsFund" in name.attrib,
        withdrew_entity=node.find("CtryNm").text.strip(),
        withdrawal_date=ApproxTimeSpan.from_str(node.find("WthdrwlDt").text),
    )


def _group_entities(entries: List[Dict]) -> CurrencyInfo:
    entities = frozenset(e["entity"] for e in entries if "entity" in e)
    withdrew_entities = (
        Historic(e["withdrew_entity"], e["name"], e["withdrawal_date"])
        for e in entries
        if "withdrew_entity" in e
    )
    withdrew_entities = tuple(sorted(withdrew_entities, key=lambda e: e.time))
    currency = entries[0]
    return CurrencyInfo(
        code=currency["code"],
        number=currency["number"],
        name=currency["name"],
        subunit_exp=currency["subunit_exp"],
        is_fund=currency["is_fund"],
        entities=entities,
        withdrew_entities=withdrew_entities,
    )
